#!/bin/env python
import gphoto2 as gp
import threading
import sys
import time 
import logging
logging.basicConfig(format='%(levelname)s: %(name)s: %(message)s', level=logging.ERROR)


class CameraControl(object):
  def __init__(self, message_queue):
    self.camera_lock = threading.Lock()
    self.status_lock = threading.Lock()
    gp.check_result(gp.use_python_logging())
    #gp.check_result(gp.gp_log_add_func(gp.GP_LOG_VERBOSE, self.log_callback))
    self.context = gp.Context()
    gp.check_result(gp.use_python_logging())
    self.cameras = {}
    self.message_queue = message_queue


  def log_callback(level, domain, string, data=None):
        print('GPhoto: level =', level, ', domain =', domain, ', string =', string)
        
  def list_cameras(self):
      if self.camera_lock.locked():
        return self.cameras
      else:
        cameras, messages = self.get_cameras(self.cameras)
        for message in messages:
          self.message_queue.put(message)
        self.cameras = cameras
        return cameras
  
    
  def __exit__(self):
    for camera, _ in self.cameras.iteritems():      
      gp.check_result(gp.gp_camera_exit(camera, self.context))

  def set_camera_config(self, camera_name, config_name, config_value):
    if camera_name in self.cameras:
      with self.camera_lock:
        camera_config = gp.check_result(gp.gp_camera_get_config(self.cameras[camera_name]["ref"], self.context))
        self.set_config(camera_config, config_name, config_value)
        ret = gp.check_result(gp.gp_camera_set_config(self.cameras[camera_name]["ref"], camera_config, self.context))
      if ret == 0:
        self.message_queue.put({"message": "%s set successfully to %s on %s"%(config_name, config_value, camera_name), "type": "success"})
      else:
        self.message_queue.put({"message": "Error while setting %s to %s on %s"%(config_name, config_value, camera_name), "type": "danger"})
      return ret == 0

  def calibrate_camera(self, camera):
    #TODO
    #self.cameras[camera]

    pass

  def set_camera_clock(self, camera):
    pass

  def get_cameras(self, cameras):
    found = []
    messages = []
    cameras_old = cameras
    with self.camera_lock:
      for name, addr in self.context.camera_autodetect():
          if name in cameras.keys() and cameras[name]["addr"] == addr:
            found += [name,]
            #print name, "found, skipping"
            camera_config = gp.check_result(gp.gp_camera_get_config(cameras[name]["ref"], self.context))
            camera_config_dict = self.get_config(camera_config)
            cameras[name]["config"] = camera_config_dict
            continue
          else:
            print "Adding", name, addr
            messages += [{"message": "Added %s"%(name), "type": "success"}]
            found += [name,]
            camera = gp.Camera()
            # search ports for camera port name
            port_info_list = gp.PortInfoList()
            port_info_list.load()
            idx = port_info_list.lookup_path(addr)
            camera.set_port_info(port_info_list[idx])
            camera.init(self.context)
            camera_config = gp.check_result(gp.gp_camera_get_config(camera, self.context))
            camera_config_dict = self.get_config(camera_config)
            cameras[name] = {"ref":camera, "config":camera_config_dict, "addr":addr}
          gp.check_result(gp.gp_camera_exit(cameras[name]["ref"], self.context))
    todelete = []
    if cameras_old:
      for name,_ in cameras_old.iteritems():
        if name not in found:
          print "Deleting", name
          messages += [{"message": "%s removed"%(name), "type": "danger"}]
          todelete += [name,]
          
    cameras = {k:v for k,v in cameras.iteritems() if k not in todelete}
    return cameras, messages
    

  def get_config(self, section):
    child_count = section.count_children()
    ret = {}
    for n in range(child_count):
        child = section.get_child(n)
        label = '{} ({})'.format(child.get_label(), child.get_name())
        child_type = child.get_type()
        if child_type == gp.GP_WIDGET_SECTION:
          #print label
          #print "---"
          child_ret = self.get_config(child)
          if len(child_ret) > 0:
            ret[label] = child_ret
        elif child_type == gp.GP_WIDGET_TEXT:
          value = child.get_value()
          if value:
            if sys.version_info[0] < 3:
                value = value.decode('utf-8')
            #print "text",label, value, []
            ret[label] = ["text", value, child.get_readonly()]
        elif child_type == gp.GP_WIDGET_RANGE:
          lo, hi, inc = child.get_range()
          value = child.get_value()
          #choices = range(int(lo * inc), int(hi * inc))
          value = (int(value * inc))
          #print "[RANGE]",label, value
          #print "Unimplemented RANGE property", label, value
        elif child_type == gp.GP_WIDGET_TOGGLE:
              value = child.get_value()
              #print "[CHECKBOX]",label, value
              ret[label] = ["checkbox", value, child.get_readonly()]
        elif child_type == gp.GP_WIDGET_RADIO or child_type == gp.GP_WIDGET_MENU:
              choice_count = child.count_choices()
              choices = []
              for n in range(choice_count):
                  choice = child.get_choice(n)
                  if choice:
                      choices += [choice]
              value = child.get_value()
              #print "[SELECT]",label, value, choices
              ret[label] = ["select", value, choices, child.get_readonly()]
        elif child_type == gp.GP_WIDGET_DATE:
            value = child.get_value()
            #print "date",label, value
            ret[label] = ["date", value, child.get_readonly()]
    return ret

  def set_config(self, section, name, value):
    child_count = section.count_children()
    for n in range(child_count):
        child = section.get_child(n)
        label = '{} ({})'.format(child.get_label(), child.get_name())

        child_type = child.get_type()
        if child_type == gp.GP_WIDGET_SECTION:
          if self.set_config(child, name, value):
            return True
        else:
          #print "Trying %s to %s"%(label, value)
          if name in label:
            print "Setting %s to %s"%(label,value)
            if child_type == gp.GP_WIDGET_TEXT:

              if sys.version_info[0] < 3:
                  value = unicode(value).encode('utf-8')
              else:
                  value = str(value)
              child.set_value(value)
              return True
            elif child_type == gp.GP_WIDGET_RANGE:

              lo, hi, inc = child.get_range()
              value = float(value) * inc
              child.set_value(value)
              return True
            elif child_type == gp.GP_WIDGET_TOGGLE:

              child.set_value((0, 1)[value])
              return True
            elif child_type == gp.GP_WIDGET_RADIO or child_type == gp.GP_WIDGET_MENU:
              choice_count = child.count_choices()
              choices = []
              for n in range(choice_count):
                  choice = child.get_choice(n)
                  if choice:
                      choices += [choice]
              print choices
              if value in choices:
                child.set_value(value)
                return True
              else:
                print "Unknown value %s for %s"%(value, label)

            elif child_type == gp.GP_WIDGET_DATE:
              
              
              child.set_value(int(value))
              return True
            else:
              return False


  def capture_image(self, camera_name, project_target_raw):
    with self.camera_lock:
      try:
        print "%s Lock acquired"%(camera_name)
        camera = self.cameras[camera_name]["ref"]
        file_path = gp.check_result(gp.gp_camera_capture(camera, gp.GP_CAPTURE_IMAGE, self.context))
        camera_file = gp.check_result(gp.gp_camera_file_get(camera, file_path.folder, file_path.name, gp.GP_FILE_TYPE_NORMAL, self.context))
        print "Output %s to %s"%(camera_file, project_target_raw)
        ret = gp.check_result(gp.gp_file_save(camera_file, project_target_raw))
      except gp.GPhoto2Error as e:
        print e
        print "Capture failed"
        return False
    print "Capture done"
    return ret == 0


    #gp.check_result(gp.gp_camera_exit(camera, self.context))
