#!/usr/bin/env python
import base64
import datetime
from PIL import Image, ImageDraw
from io import BytesIO
import subprocess
import numpy as np
import matplotlib
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from matplotlib.figure import Figure
#matplotlib.use('Agg')
#import matplotlib.pyplot as plt
import matplotlib.image as mpimg
import os, os.path
import scipy.misc

def get_image(infile, maxsize=0, fmt='JPEG'):
  #print "Get_image", infile
  im = Image.open(infile)
  
  if maxsize > 0:
    maxsize = (maxsize,maxsize)
  else:
    maxsize = im.size
  #print "Thumbnailing"
  im.thumbnail(maxsize, Image.ANTIALIAS)
  
  #print "Thumbnailing done"
  newfp = BytesIO()
  im.save(newfp, fmt)
  newfp.seek(0)
  #print "encoding to b64"
  return base64.b64encode(newfp.read())

def thumbnail(raw_path, jpeg_path, maxsize=0):
  print "Extracting thumbnail"
  with BytesIO() as out:
    proc = subprocess.Popen(['dcraw','-h','-q','0', '-c', '-T', raw_path], stdout=subprocess.PIPE) 
    out.write(proc.communicate()[0])
    out.seek(0)
    
    im = Image.open(out)
    if maxsize > 0:
      maxsize = (maxsize,maxsize)
      im.thumbnail(maxsize, Image.ANTIALIAS)
      im.save(jpeg_path, "JPEG")
    else:
      im.save(jpeg_path, "JPEG")
      
        
    

def tiff(raw_path, tiff_path, force=False):
  if force or not os.path.isfile(tiff_path):
    print "Extracting TIFF"
    with open(tiff_path, "w") as out:
      return 0 == subprocess.call(['dcraw','-T', '-W', '-w', '-c' , raw_path], stdout=out)
  else:
    print "TIFF exists, skipping"
    return True

def metadata(raw_path):
  print "Extracting metadata"
  out = subprocess.check_output(['dcraw','-i','-v',raw_path])
  meta_lines = out.strip().split('\n')
  
  meta = {l.split(':')[0].lower():":".join(l.split(':')[1:]).strip() for l in meta_lines}
  
  
  meta["timestamp"] = datetime.datetime.strptime(meta["timestamp"], "%a %b %d %H:%M:%S %Y").isoformat()
  shutter = meta["shutter"].split(" ")[0]
  
  meta["shutter"] = shutter
  meta["iso speed"] = np.int(meta["iso speed"])
  meta["aperture"] = np.float(meta["aperture"].split("/")[1])
  return meta

def weighting_constant(iso, aperture, shutter, K_c):
  shutter = shutter.split("/")
  if len(shutter) == 2:
    shutter = np.float(np.float(shutter[0])/np.float(shutter[1]))
  else:
    shutter = np.float(shutter[0])
  return  (aperture**2.0)/(K_c*shutter*iso)
  

def get_K_c(iso, aperture, shutter, pixel_value, luminance):
  shutter = shutter.split("/")
  if len(shutter) == 2:
    shutter = np.float(np.float(shutter[0])/np.float(shutter[1]))
  else:
    shutter = np.float(shutter[0])
  print shutter, aperture, pixel_value, iso, luminance
  return (np.float(aperture)**2.0 * np.int(pixel_value))/(shutter*np.int(iso)*np.float(luminance))

    

def get_raw_pixel_data(img_path, force=False):
  npy_fname = img_path+".npz"
  if not force  and os.path.isfile(npy_fname):
    b = np.load(npy_fname)
    ret = b.items()[0][1]
    b.close()
    return ret, ret.min(), ret.max()
  else:
    im = Image.open(img_path)
    gim = im.convert('L')
    del im    
    ngim = np.asarray(gim, dtype=np.uint8)
    np.savez(img_path, ngim)
    return ngim, ngim.min(), ngim.max()
  
def luminance_bar(target_img_path, min_pix, max_pix,  height=25.0, width=600, pixel_data=None):
  fig = Figure()
  dpi=fig.get_dpi()
  fig.set_size_inches(width/float(dpi),height/float(dpi))
  canvas = FigureCanvas(fig)
  ax = fig.add_subplot(111)
  cmap = matplotlib.cm.spectral
  norm = matplotlib.colors.Normalize(vmin=min_pix, vmax=max_pix)
  cb = matplotlib.colorbar.ColorbarBase(ax, cmap=cmap, norm=norm, orientation='horizontal')
  lut = np.around(cmap._lut[:256,:3]*255).astype(np.uint8)
  l = {}
  l["r"] = lut[:,0].tolist()
  l["g"] = lut[:,1].tolist()
  l["b"] = lut[:,2].tolist()
  l["vals"] = cb._values.tolist()
  ax.set_xlabel("cd/m^2")
  fig.savefig(target_img_path, bbox_inches='tight', pad_inches=0)
  return l

def luminance(source_img_path, target_img_path, weighting_constant, maxsize=0, force=False, pixel_data=None):
  
  if isinstance(pixel_data, type(None)):
    pixel_data, pixel_data_min, pixel_data_max = get_raw_pixel_data(source_img_path, force)
  gim = Image.fromarray((pixel_data*np.float32(weighting_constant)).astype(np.float32))

  if maxsize > 0:
    maxsize = (maxsize, maxsize)
  else:
    maxsize = gim.size
  gim.thumbnail(maxsize, Image.ANTIALIAS)
  print "Size BW weighted - thumbnail:", gim.size
  matplotlib.image.imsave(target_img_path, np.array(gim), cmap=matplotlib.cm.spectral)
  print "Saved"

def crop_rect(img, rect):
  (x,y,w,h) = rect
  poly_mask = np.zeros_like(img).astype(bool)
  print "Original shape: ", poly_mask.shape
  
  if y+h > poly_mask.shape[0]:
    print "Correcting Height"
    h = poly_mask.shape[0]-y
    
  
  if x+w > poly_mask.shape[1]:
    print "Correcting Width"
    w = poly_mask.shape[1]-x
    
  poly_mask[y:y+h, x:x+w] = True
  
  crop = img[np.where(poly_mask)]
  print "Crop shape: ", h, w
  print "Crop max: ", x+w,y+h
  print crop.shape[0]/np.float(w), crop.shape[0]/np.float(h)
  print "reshape to shape: ", (h, w)
  return crop.reshape(h, w)

  

def histogram(source_img_path, target_img_path, weighting_constant=1.0, rect=None, force=False, pixel_data=None):
  if isinstance(pixel_data, type(None)):
    pixel_data, pixel_data_min, pixel_data_max = get_raw_pixel_data(source_img_path, force)
  if rect:
    pixel_data = crop_rect(pixel_data, rect)
  print "Creating histogram"
  
  fig = Figure(figsize=(9, 4.5))
  canvas = FigureCanvas(fig)
  ax = fig.add_subplot(111)
  percentage=15
  h = ax.hist(scipy.misc.imresize(pixel_data, percentage).flatten()*weighting_constant, 256, facecolor='black', normed=True)
  #print pixel_data.shape, scipy.misc.imresize(pixel_data, 50).shape
  #fig.gca().axes.get_xaxis().set_visible(False)
  fig.gca().axes.get_yaxis().set_visible(False)
  #fig.gca().set_xlim([0,255])
  fig.savefig(target_img_path, bbox_inches='tight', pad_inches=0)

def get_polygon_points(x,y,w,h):
  return [(x,y), (x+w,y), (x+w,y+w), (x, y+w)]
  
def polygon_stats(source_img_path, constant, rect, force=False, pixel_data=None):
  print "Computing rectangular stats", rect
  if isinstance(pixel_data, type(None)):
    pixel_data, pixel_data_min, pixel_data_max = get_raw_pixel_data(source_img_path, force)  
  crop = crop_rect(pixel_data, rect)
  percentage = 15
  #print crop.min(), scipy.misc.imresize(crop, percentage).min(), crop.max(), scipy.misc.imresize(crop, percentage).max(), crop.mean(), scipy.misc.imresize(crop, percentage).mean()
  crop = scipy.misc.imresize(crop, percentage)
  del pixel_data
  ret = dict()
  
  ret['min'] = str((crop*constant).min().round(decimals=2))
  ret['max'] = str((crop*constant).max().round(decimals=2))
  ret['mean'] = str((crop*constant).mean().round(decimals=2))
  return ret

def get_line_stats(source_img_path, weighting_constant, target_img_path, line, force=False, pixel_data=None):
  if isinstance(pixel_data, type(None)):
    x1,y1,x2,y2 = line
    idxs = get_line_idxs(x1,y1,x2,y2)
    pixel_data = get_raw_pixel_data(source_img_path, force)[0][idxs[:,0], idxs[:,1]]
  
  print "Creating line histogram"
  #print ngim.shape, ngim.min(), ngim.max(), ngim.mean()
  fig = Figure(figsize=(9, 4.5))
  canvas = FigureCanvas(fig)
  ax = fig.add_subplot(111)
  h = ax.hist(pixel_data.flatten()*weighting_constant, 256, facecolor='black', normed=True)
  #fig.gca().axes.get_xaxis().set_visible(False)
  fig.gca().axes.get_yaxis().set_visible(False)
  #fig.gca().set_xlim([0,255])
  fig.savefig(target_img_path, bbox_inches='tight', pad_inches=0)
  ret = dict()
  ret['min'] = str((pixel_data*weighting_constant).min().round(decimals=2))
  ret['max'] = str((pixel_data*weighting_constant).max().round(decimals=2))
  ret['mean'] = str((pixel_data*weighting_constant).mean().round(decimals=2))
  return ret
  
  
def write_report_img(img,out, line=None, point=None, rect=None):
  im = Image.open(img)
  draw = ImageDraw.Draw(im)
  if line:
    draw.line(line, fill=0, width=10)
  if point:
    px, py = point
    size = 30
    draw.line((px-size, py-size, px+size, py+size), fill=0, width=10)
    draw.line((px+size, py-size, px-size, py+size), fill=0, width=10)
  del draw
  if rect:
    print im.size
    im = im.crop(rect)
  im.save(out)
  

def get_directive(x1,y1, x2, y2):
  return np.linalg.solve(np.array([[1,x1],[1,x2]]), np.array([y1,y2]))

def get_line_idxs(x1,y1,x2,y2):
  q, k = np.linalg.solve(np.array([[1,x1],[1,x2]]), np.array([y1,y2]))
  f = lambda x: np.int(np.ceil(x*k + q))
  if x1 > x2:
    idxs = np.array([[f(x), int(x)] for x in np.arange(x2,x1+1)])
  else:
    idxs = np.array([[f(x), int(x)] for x in np.arange(x1,x2+1)])
  return idxs
  

