#!/usr/bin/env python
import threading
import image_analysis
import os, errno
import time
import utils



class ProjectWorker(threading.Thread):
  def __init__(self, capture_queue,message_queue, camera_control, K_c,  project_root, projectid=None, project_config={}, seconds_to_start=0):
    threading.Thread.__init__(self)
    self.status_lock = threading.Lock()
    self.project_config = project_config
    self.terminate_project = False
    self.project_running = False
    self.ncaptures = project_config["ncaptures"]
    self.capture_interval = project_config["capture_interval"]
    self.cameras = project_config["cameras"]
    self.capture_thread_ref = []
    self.capture_queue = capture_queue
    self.camera_control = camera_control
    self.K_c = K_c
    self.message_queue = message_queue
    self.project_id = projectid
    self.wakeup_interval = 1.0
    self.project_config["project_id"] = projectid
    self.project_root = project_root
    self.seconds_to_start = int(seconds_to_start)
    
      
  
  def run(self):
    self.project_running = True
    self.ncaptures_todo = self.ncaptures
    project_paths = {camera: str("%s/%s"%(self.project_root,utils.get_camera_path(camera))) for camera in self.cameras}
    for path in project_paths.values():
      print path
      utils.mkdir_p(path)
    
    if self.seconds_to_start > 0:
      self.capture_queue.put({"projectid": self.project_id, "data": {"waiting":  True}, "status": "OK"})
      print "Waiting %d seconds for scheduled start"%(self.seconds_to_start)
      self.interruptable_sleep(self.seconds_to_start)
    self.capture_queue.put({"projectid": self.project_id, "data": {"waiting":  False}, "status": "OK"})
    for i in xrange(self.ncaptures):
      if self.terminate_project:
        print "Terminating project %s"%(self.project_id)
        for t in self.capture_thread_ref:
          t.join()
        print "Project %s terminated"%(self.project_id)
        break
      print "Capturing %d image"%(i)
      self.capture_thread_ref = []
      for camera in self.cameras:
        tr = Capture(self.capture_queue, self.camera_control, camera, self.project_id, i+1, self.K_c[camera], self.message_queue, project_paths[camera])
        tr.start()
        self.capture_thread_ref += [tr]
      self.ncaptures_todo -= 1

        
      self.capture_queue.put({"projectid": self.project_id, "data": {"ncaptures_todo":  self.ncaptures_todo}, "status": "OK"})
      if self.ncaptures_todo > 0:
        self.interruptable_sleep(self.capture_interval)
    for t in self.capture_thread_ref:
      t.join()
    self.project_running = False
    print "Ending project %s"%(self.project_id)
    self.capture_queue.put({"projectid": self.project_id, "data": {"finished": True, "status": "OK"}})
    self.message_queue.put({"message": "Project finished", "type": "success"})
  
  def stop_project(self):
    if self.project_running:
      self.terminate_project = True
      print "Waiting for capture thread to die"
      self.thread_ref.join()
      self.terminate_project = False
      self.project_running = False
      self.thread_ref = None
      
  def interruptable_sleep(self, nseconds):
    if nseconds > self.wakeup_interval:
      for nsec in xrange(nseconds):
        if self.terminate_project:
          break
        self.capture_queue.put({"projectid": self.project_id, "data": {"ncaptures_countdown":  nseconds-nsec}, "status": "OK"})
        time.sleep(self.wakeup_interval)  

class Capture(threading.Thread):
    def __init__(self, capture_queue, camera_control, camera, project_id, i, K_c, message_queue, project_root):
      threading.Thread.__init__(self)
      self.project_root = project_root
      self.message_queue = message_queue
      self.capture_queue = capture_queue
      self.camera = camera
      self.camera_control = camera_control
      self.project_id = project_id
      self.i = i
      self.K_c = K_c
      self.project_target = "%s/%d"%(self.project_root, i)
      self.project_target_raw = "%s/%d.raw"%(self.project_root, i)
      self.project_target_tiff = "%s/%d.tiff"%(self.project_root, i)
      self.project_target_jpeg = "%s/%d.jpeg"%(self.project_root, i)
      self.project_target_bw_tiff = "%s/%d.bw.tiff"%(self.project_root, i)
      self.project_target_luminace = "%s/%d.luminance.png"%(self.project_root, i)
      self.project_target_histogram = "%s/%d.histogram.png"%(self.project_root, i)

    def postprocess(self):
      metadata = image_analysis.metadata(self.project_target_raw)
      self.capture_queue.put({"projectid": self.project_id,"camera": self.camera, "captureid": self.i, "data": {"metadata": metadata}, "status": "OK"})
      image_analysis.thumbnail(self.project_target_raw, self.project_target_jpeg, maxsize=1024)
      b64_img = image_analysis.get_image(self.project_target_jpeg, fmt='JPEG')
      self.capture_queue.put({"projectid": self.project_id,"camera": self.camera, "captureid": self.i, "data": {"thumbnail": {"path": self.project_target_jpeg, "data": b64_img}}, "status": "OK"})
      
      
      weighting_constant = image_analysis.weighting_constant(metadata["iso speed"], metadata["aperture"], metadata["shutter"], self.K_c)
      
      image_analysis.histogram(self.project_target_jpeg, self.project_target_histogram, weighting_constant)
      b64_img = image_analysis.get_image(self.project_target_histogram, fmt='JPEG')
      self.capture_queue.put({"projectid": self.project_id,"camera": self.camera, "captureid": self.i, "data": {"histogram": {"path": self.project_target_histogram, "data": b64_img}}, "status": "OK"})
      
      image_analysis.luminance(self.project_target_jpeg, self.project_target_luminace, weighting_constant)
      b64_img = image_analysis.get_image(self.project_target_luminace, fmt='JPEG')
      self.capture_queue.put({"projectid": self.project_id,"camera": self.camera, "captureid": self.i, "data": {"luminance":  {"path": self.project_target_luminace, "data": b64_img}}, "status": "OK"})
      
      capture_data = {"status":True,"id": self.i,"project_id":self.project_id,"camera_name":self.camera,"timestamp":metadata["timestamp"],"iso":metadata["iso speed"],"shutter":metadata["shutter"],"aperture":metadata["aperture"]}
      return capture_data
      

    def run(self):
      start = time.time()
      print self.project_target_raw
      if self.camera_control.capture_image(self.camera, self.project_target_raw):
        self.message_queue.put({"message": "Capture %d done [%s]"%(self.i, self.camera), "type": "success"})
        self.capture_queue.put({"projectid": self.project_id,"camera": self.camera, "captureid": self.i, "data": {"capture": True, "status": "OK"}})
        try:
          capture_data = self.postprocess()
          self.capture_queue.put({"projectid": self.project_id,"camera": self.camera, "captureid": self.i, "data": {"model": capture_data}, "status": "OK"})
          self.message_queue.put({"message": "Postprocessing %d done"%(self.i), "type": "success"})
        except Exception as e:
          print e
          self.message_queue.put({"message": "Error in  postprocessing %d"%(self.i), "type": "danger"})
          self.message_queue.put({"message": "Exception: %s"%(str(e)), "type": "debug"})
        self.message_queue.put({"message": "Capture took %fs"%(time.time() - start), "type": "debug"})
      else:
        capture_data = {"status":False,"id": self.i,"project_id":self.project_id,"camera_name":self.camera}
        self.capture_queue.put({"projectid": self.project_id,"camera": self.camera, "captureid": self.i, "data": {"model": capture_data}, "status": "OK"})
        self.capture_queue.put({"projectid": self.project_id,"camera": self.camera, "captureid": self.i, "data": {"capture": False, "status": "OK"}})
        self.message_queue.put({"message": "Error in %d capture [%s]"%(self.i, self.camera), "type": "danger"})
